/******** <lalVerbatim file="BlockApplySearchCV"> ********
Author: Tibbits, M. M.
$Id: BlockApplySearch.c,v 1.2 2002/07/05 20:44:36 tibbits Exp $
********* </lalVerbatim> ********/

/******************************* <lalLaTeX file="BlockApplySearchC">
\subsection{Module \texttt{BlockApplySearch.c}}
\label{s:BlockApplySearch.c}

Routine that uses parallelism of processes to implement the rho2 backround
search.  The rho2 search is based on, a yet to be published, block
normailization paper.  Our search returns a result or "expectation" value
which corresponds to the probability of there being multiple gaussian
waveforms within the measured data.  The search also returns the index
and magnitude of the maximum value in the input vector.  Via a filterparam
you can also enable the Rho3 search which chops the input at the maximum
value and reruns the Rho2 code on both segements.  Thereby, we can search
for both the beginning and the end of a gaussian waveform present.

\subsubsection*{Prototypes}
\input{BlockApplySearchCP}
\index{\texttt{LALApplySearch()}}

\subsection*{Description}

Within the LALApplySearch Routine, the block search master receives the
first set of results from the slaves (processed within the LALApplySearch
on the slave node).  Then the master node either sends a data id of -1
signalling that all vectors that need processing have been sent out or the
master node send the index of the vector stored within the master's parameter
structure in a vector sequence "data".

On each slave node, as in the LALConditionData section, the slave first
processes the data vector that it received in LALConditionData.  Next the 
slave receives the id of the next vector and either exits if -1 is given or
waits to receive the next data vector if signalled by the master node.  At
the completetion of Rho2, if the Rho3 flag is set, each slave will wait for
the next input vecto to process.

The result values are stored in an array on the master node until all vectors
have been processed.  The results are then built for subsequent insertion
into the metadata database and frames using the \texttt{BlockBuildOutput}
routine.

\subsection*{Algorithms}

The LALApplySearch routine does not manipulate the data in any way on either
master or the slave until the slave sends the data to the Rho2 routine.  The
LALApplySearch routine is designed as a wrapper to the facilitate 
parallelism mentioned earlier.  The Block Normal Search takes an input of a
vector sequence that can have any number of vectors in it.  The parallelism
begins by breaking the input into individual vectors and sending each
vector to a different slave node.  Each slave node then preforms the Rho2
algorithm on its vector and returns the results

To note, the Rho2 algorithm is discussed below under the \texttt{BlockRho2}
routine heading.

The MPI exchange is as follows:
%
\begin{enumerate}
\item 
The Master waits to receive a message from a slave node that the slave
is ready to send its data.

\item 
The Slave then sends first the index of the data vector that it
processed.

\item 
Then the Slave sends its results

\item
The Master then sends back the index of the next vector to be processed
which is set to -1 if all data vectors have been processed.

\item
The Master then sends the next data vector to the slave, given that
there are still vectors to be processed.

\end{enumerate}

\subsection*{Uses}
 
\begin{itemize}
\item \texttt{LALMPISendMsg}
\item \texttt{LALMPIRecvMsg}
\item \texttt{LALMPISendINT4}
\item \texttt{LALMPIRecvINT4}
\item \texttt{LALMPISendREAL8}
\item \texttt{LALMPIRecvREAL8}
\item \texttt{LALDCreateVector}
\item \texttt{LALMPISendREAL8Vector}
\item \texttt{LALMPIRecvREAL8Vector}
\item \texttt{BlockRho2}
\item \texttt{BlockBuildOutput} 
\item \texttt{LALDDestroyVector}
\end{itemize}

\subsection*{Notes}

\vfill{\footnotesize\input{BlockApplySearchCV}}

 *************************************************** </lalLaTeX> */

#include "BlockApplySearch.h"
#include "BlockConditionData.h"
#include "BlockBuildOutput.h"


NRCSID( BLOCKAPPLYSEARCHC, "$Id: BlockApplySearch.c,v 1.2 2002/07/05 20:44:36 tibbits Exp $");

/************************ <lalVerbatim file="BlockApplySearchCP"> */
void LALApplySearch
(
	LALStatus		*status,
	LALSearchOutput		*output,
	LALSearchInput		*input,
	LALApplySearchParams	*applyParams
)
/*********************************************** </lalVerbatim> */
{
	/*  Variable Declarations  */
	BlockSearchParams	*params;

	INITSTATUS( status, "BlockApplySearch", BLOCKAPPLYSEARCHC );
	ATTATCHSTATUSPTR (status);

	/*  Inititalize Variables  */
	params = ((BlockSearchParams*)(applyParams->searchParams));

	/*  Sanity Checks on Input  */
	ASSERT( output, status, BLOCKAPPLYSEARCHH_ENULL, BLOCKAPPLYSEARCHH_MSGENULL);

	ASSERT( input, status, BLOCKAPPLYSEARCHH_ENULL, BLOCKAPPLYSEARCHH_MSGENULL);

	ASSERT( applyParams, status, BLOCKAPPLYSEARCHH_ENULL, BLOCKAPPLYSEARCHH_MSGENULL);

	/*  Get rank  */
	MPI_Comm_rank(*(applyParams->mpiParams->comm), &(params->rank));

	/*  Which node is this running on?  */
	if (params->searchMaster)
	{
		/* master */
		ApplySearchMaster(status->statusPtr, output, input, applyParams );
	}
	else
	{
		/* slave */
		ApplySearchSlave(status->statusPtr, output, input, applyParams );
	}

	DETATCHSTATUSPTR (status);
	RETURN(status);
}

/*
 * Function ApplySearchMaster
 *
 * Called from LALApplySearch only on the search master node. It's
 * main job is to wait for responses for each slave ie. retrieve the
 * statistics, send the next vector, wait for the slaves to end, and
 * then to format the output.
 */

/************************ <lalVerbatim file="BlockApplySearchMasterCP"> */
void ApplySearchMaster
(
	LALStatus		*status,
	LALSearchOutput		*output,
	LALSearchInput		*input,
	LALApplySearchParams	*applyParams
)
/*********************************************** </lalVerbatim> */
{
	/*  Variable Declaration  */
	BlockSearchParams	*myParams;
	REAL8Vector		*tempVector;
	MPIMessage		mpiMsg;
	MPI_Comm		*comm;
	UINT4			segLength;
	INT4			sequenceNumber;
	INT4			iterator;
	INT4			index;
	INT4			key;
	INT4			id;


	INITSTATUS( status, "ApplySearchMaster", BLOCKAPPLYSEARCHC );
	ATTATCHSTATUSPTR(status);

	/*  No input checking done here  */
	/*  All assertions in LALApplySearch Routine above  */

	/*  Variable Initialization  */
	comm		= ((MPI_Comm*)(applyParams->mpiParams->comm));
	myParams	= ((BlockSearchParams*)(applyParams->searchParams));
	sequenceNumber	= 0;
	tempVector	= NULL;
	segLength	= 0;
	index		= 0;
	key		= 0;
	id		= myParams->nextVectorToSend;

	/*  Wait until all slaves have sent their last data point  */
	if( myParams->segmentsRemaining < myParams->numberOfSequences )
	{
		/*  Receive results from Slave  */
		/*  Wait for a slave to signal it is done processing and ready to send data  */
		LALMPIRecvMsg( status->statusPtr, &mpiMsg, (*comm));
		CHECKSTATUSPTR( status );

		/*  Slave sends segment ID  */
		LALMPIRecvINT4(status->statusPtr, &(index), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Slave sends result  */
		LALMPIRecvREAL8( status->statusPtr, &(myParams->res->data[index]), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

                /*  Slave sends max value  */
                LALMPIRecvREAL8( status->statusPtr, &(myParams->value->data[index]), mpiMsg.source, (*comm));
                CHECKSTATUSPTR( status );

                /*  Slave sends index of max value  */
                LALMPIRecvINT4( status->statusPtr, &(myParams->ndx->data[index]), mpiMsg.source, (*comm));
                CHECKSTATUSPTR( status );

		/*  If all vectors processed, set termination condition  */
                if ( (id == myParams->numberOfSequences))
                        id = -1;

		/*  Send Next Vector  */
		/*  Vector id  */
                LALMPISendINT4( status->statusPtr, &(id), mpiMsg.source, (*comm));
                CHECKSTATUSPTR( status );

		/*  If not in termination state continue  */
		if ( id != -1 )
		{
			segLength = myParams->data->vector[id]->length;

			/*  Send length of vector  */
			LALMPISendUINT4( status->statusPtr, &(segLength), mpiMsg.source, (*comm));
			CHECKSTATUSPTR( status );

			/*  Create and fill temporary vector to send to Slave node  */
			LALDCreateVector(status->statusPtr, &tempVector,  myParams->data->vector[id]->length);
			CHECKSTATUSPTR( status );

			for( iterator = 0; iterator < tempVector->length; iterator++)
			{
				tempVector->data[iterator] = myParams->data->vector[id]->data[iterator];
			}

			/*  Send Data Vector  */
			LALMPISendREAL8Vector(status->statusPtr, tempVector, mpiMsg.source, (*comm));
			CHECKSTATUSPTR( status );

			/*  Clean up temporary Mem each time used  */
			LALDDestroyVector(status->statusPtr, &tempVector);
			CHECKSTATUSPTR( status );
			tempVector	= NULL;
		}

		/*  If not in termination, then one of the slaves just  */
		/*  got the current vector myParams->data->vector[id]->data ~ see above for exact algorithm  */
		if (id != -1)
			id++;

		output->fracRemaining	= ( 1 - myParams->segmentsRemaining / myParams->numberOfSequences );
		output->notFinished 	= 1;

		myParams->segmentsRemaining++;
	}
	else
	{
		if ( myParams->reprocessWithRho3 )
		{
			id = 0;

			printf("myParams->reprocessWithRho3 = %d\n",myParams->reprocessWithRho3);

			BlockReparseData( status->statusPtr, myParams );

			if ( !myParams->dataRho3 )
			{
				id = -1;
				myParams->reprocessWithRho3--;
			}
			else
			{
				if ( myParams->data != NULL)
				{
					for ( iterator = 0; iterator < myParams->data->numberOfVectors; iterator++)
					{
						LALDDestroyVector(status->statusPtr, &(myParams->data->vector[iterator]));
						CHECKSTATUSPTR(status);
					}

					LALFree(myParams->data->vector);
					CHECKSTATUSPTR(status);

					LALFree(myParams->data);
					CHECKSTATUSPTR(status);

					myParams->data = myParams->dataRho3;
				}

				myParams->segmentsRemaining = 0;

				for ( iterator = 0; iterator < myParams->res->length; iterator++)
				{
					if ( myParams->res->data[iterator] > myParams->threshold )
					{
						AddResults( status->statusPtr, output, iterator, myParams );
						myParams->numOutput++;
					}
				}

				myParams->run++;

				LALDDestroyVector(status->statusPtr, &(myParams->res));
				CHECKSTATUSPTR(status);

				LALDDestroyVector(status->statusPtr, &(myParams->value));
				CHECKSTATUSPTR(status);

				LALI4DestroyVector(status->statusPtr, &(myParams->ndx));
				CHECKSTATUSPTR(status);

				LALDCreateVector( status->statusPtr, &(myParams->res), myParams->numberOfSequences);
				CHECKSTATUSPTR ( status );

				LALDCreateVector( status->statusPtr, &(myParams->value), myParams->numberOfSequences);
				CHECKSTATUSPTR ( status );

				LALI4CreateVector( status->statusPtr, &(myParams->ndx), myParams->numberOfSequences);
				CHECKSTATUSPTR ( status );

				myParams->reprocessWithRho3--;
			}

			for ( sequenceNumber = 0; sequenceNumber < myParams->numSlaves; sequenceNumber++)
			{
				/*  Vector id  */
				LALMPISendINT4( status->statusPtr, &(id),  sequenceNumber + 1, (*comm));
				CHECKSTATUSPTR( status );

				/*  If not in termination state continue  */
				if ( id != -1 )
				{
					segLength = myParams->dataRho3->vector[id]->length;

					/*  Send length of vector  */
					LALMPISendUINT4( status->statusPtr, &(segLength),  sequenceNumber + 1, (*comm));
					CHECKSTATUSPTR( status );

					/*  Create and fill temporary vector to send to Slave node  */
					LALDCreateVector(status->statusPtr, &tempVector,  myParams->dataRho3->vector[id]->length);
					CHECKSTATUSPTR( status );

					for( iterator = 0; iterator < tempVector->length; iterator++)
					{
						tempVector->data[iterator] = myParams->dataRho3->vector[id]->data[iterator];
					}

					/*  Send Data Vector  */
					LALMPISendREAL8Vector(status->statusPtr, tempVector,  sequenceNumber + 1, (*comm));
					CHECKSTATUSPTR( status );

					/*  Clean up temporary Mem each time used  */
					LALDDestroyVector(status->statusPtr, &tempVector);
					CHECKSTATUSPTR( status );
					tempVector      = NULL;

					if ( id < myParams->numberOfSequences )
						id += 1;
					else
						id = -1;
				}
			}

			myParams->dataRho3 = NULL;
		}
		else
		{
			/*  Format the result array into an output object to fill the database  */
			BuildOutput( status->statusPtr, output, myParams );

			output->fracRemaining	= 0;
			output->notFinished	= 0;

		}
	}

	myParams->nextVectorToSend = id;

	/*  Normal Exit  */
	DETATCHSTATUSPTR(status);
	RETURN(status);
}


/*
 * Function ApplySearchSlave
 *
 * Called from LALApplySearch on each slave node. It receives the data
 * sent by the master node, processes it via the BlockRho2 function,
 * then it passes the results to the master node, and if there are more
 * vectors to be processed, it waits to receive the next one.
 */


/************************ <lalVerbatim file="BlockApplySearchMasterCP"> */
void ApplySearchSlave
(
	LALStatus		*status,
	LALSearchOutput		*output,
	LALSearchInput		*input,
	LALApplySearchParams	*applyParams
)
/*********************************************** </lalVerbatim> */
{
	/*  Variable Declaration  */
	BlockSearchParams	*myParams;
	REAL8Vector		*local;
	MPIMessage		sendMsg;
	MPI_Comm		*comm;
	REAL8			sum;
	REAL8			rpeak;
	INT4			index;

	INITSTATUS( status, "ApplySearchSlave", BLOCKAPPLYSEARCHC );
	ATTATCHSTATUSPTR(status);

        /*  No input checking done here  */
        /*  All assertions in LALApplySearch Routine above  */

	/*  Variable Initialization  */
	comm		= ((MPI_Comm*)(applyParams->mpiParams->comm));
	myParams	= ((BlockSearchParams*)(applyParams->searchParams));
	sendMsg.source	= myParams->rank;
	sendMsg.msg	= PassData;
	sendMsg.send	= 1;
	sum		= 0;
	rpeak		= 0;
	index		= 0;
	local		= NULL;

	if ( myParams->vectorId >= 0)
	{
		/*  Send current data vector myParams->vec to BlockRho2 Function  */
		BlockRho2(status->statusPtr, &sum, &rpeak, &index, myParams->vec);
		CHECKSTATUSPTR( status );

		/*  Print a listing of the results  */
		printf("result[%d] = %3.16f\n", myParams->vectorId, (REAL4)(log(sum)));
		printf("rpeak[%d] = %3.16f\n", myParams->vectorId, (REAL4)(log(rpeak)));

		/*  Send Message to say I am ready to send */
		LALMPISendMsg( status->statusPtr, &sendMsg, 0, (*comm));
		CHECKSTATUSPTR( status );	

		/*  Send index of input data vector  */
		LALMPISendINT4( status->statusPtr, &(myParams->vectorId), 0, (*comm));
		CHECKSTATUSPTR( status );

		/*  Here comes the Result  */
		LALMPISendREAL8( status->statusPtr, &sum, 0, (*comm));
		CHECKSTATUSPTR (status);

		/*  Here comes the Max Value  */
		LALMPISendREAL8( status->statusPtr, &rpeak, 0, (*comm));
		CHECKSTATUSPTR (status);

		/*  Here comes the Index of Max Value  */
		LALMPISendINT4( status->statusPtr, &index, 0, (*comm));
		CHECKSTATUSPTR (status);

		/*  Clean up  */
		/*  This step also kills &local as both point to the same allocation  */
		LALDDestroyVector(status->statusPtr, &myParams->vec);
		CHECKSTATUSPTR(status);

		/*  Receive next index  */
		LALMPIRecvINT4( status->statusPtr, &(myParams->vectorId), 0, (*comm));
		CHECKSTATUSPTR (status);
	}

	/*  We're done, but do we need to reprocess?  */

	if ( myParams->vectorId == -1)
	{
		if ( myParams->reprocessWithRho3 )
		{
			/*  Receive next index  */
			LALMPIRecvINT4( status->statusPtr, &(myParams->vectorId), 0, (*comm));
			CHECKSTATUSPTR (status);
			myParams->reprocessWithRho3--;
		}
		else
		{
			output->numOutput	= 0;
			output->result		= NULL;
			output->fracRemaining	= 0;
			output->notFinished	= 0;
		}
	}

	if ( myParams->vectorId != -1 )
	{
		local = NULL;

		/*  How much Data?  */
		LALMPIRecvUINT4( status->statusPtr, &(myParams->segmentLength), 0, (*comm));
		CHECKSTATUSPTR (status);

		/*  Create new structure to hold local data  */
		LALDCreateVector(status->statusPtr, &local, myParams->segmentLength );
		CHECKSTATUSPTR (status);

		/*  Receive input data  */
		LALMPIRecvREAL8Vector(status->statusPtr, local, 0, (*comm));
		CHECKSTATUSPTR (status);

		/*  Set the input vector to the newly received data  */
		myParams->vec = local;
	}

	DETATCHSTATUSPTR( status );
	RETURN(status);
}


/************************ <lalVerbatim file="BlockParseDataCP"> */
void BlockReparseData
(
	LALStatus	*status,
	void		*searchParams
)
/*********************************************** </lalVerbatim> */
{

	/*  Variable Declarations  */
	BlockSearchParams	*params;
	REAL8VectorVector	*tempVector;
	UINT4			sequenceNumber;
	UINT4			iterator;
	UINT4			k;
	INT4			count;

	/*  Initialize status to pass to other functions  */
	INITSTATUS( status, "BlockReparseData", BLOCKAPPLYSEARCHC );
	ATTATCHSTATUSPTR (status);

	/*  Variable Initialization  */
	params		= ((BlockSearchParams*)(searchParams));
	tempVector	= params->dataRho3 = NULL;
	count		= 0;

	/*  Must protect against smallest vector allowed  ( vector must have 6 points )  */
	for ( iterator = 0; iterator < params->ndx->length; iterator++)
		if ( params->res->data[iterator] > params->threshold )
			count+= 2;

	if ( count )
	{
		tempVector = (REAL8VectorVector*)(LALMalloc(sizeof(REAL8VectorVector)));
	}
	else
	{
		DETATCHSTATUSPTR (status);
		RETURN(status);
	}

	if( !tempVector )
	{
		ABORT (status, BLOCKAPPLYSEARCHH_EALOC, BLOCKAPPLYSEARCHH_MSGEALOC);
	}

        params->numberOfSequences = tempVector->numberOfVectors = count;

	tempVector->vector = (REAL8Vector**) LALMalloc(sizeof(REAL8Vector*) * count);
	if( !tempVector->vector)
	{
		ABORT (status, BLOCKAPPLYSEARCHH_EALOC, BLOCKAPPLYSEARCHH_MSGEALOC);
	}

	iterator = 0;

	for ( sequenceNumber = 0; sequenceNumber < params->data->numberOfVectors; sequenceNumber++)
	{
		if ( params->res->data[sequenceNumber] > params->threshold )
		{
			/*  First Half  */
			tempVector->vector[iterator] = NULL;
			count = params->ndx->data[sequenceNumber] + 1;

			if ( count < 5 )
			{
				count = 5;
			}
			else
			{
				if ( ( params->data->vector[sequenceNumber]->length - count ) < 5 )
				{
					count = params->data->vector[sequenceNumber]->length - 5;
				}
			}


			LALDCreateVector(status->statusPtr, &(tempVector->vector[iterator]), count);
			CHECKSTATUSPTR (status);

			for ( k = 0; k < count; k++)
			{
				tempVector->vector[iterator]->data[k] = ((REAL8)(params->data->vector[sequenceNumber]->data[k]));
			}

			iterator++;


			/*  Second Half  */
			tempVector->vector[iterator] = NULL;
			LALDCreateVector(status->statusPtr, &(tempVector->vector[iterator]), (params->data->vector[sequenceNumber]->length - count));
			CHECKSTATUSPTR (status);

			for ( k = count; k < params->data->vector[sequenceNumber]->length; k++ )
			{
				tempVector->vector[iterator]->data[k - count] = ((REAL8)(params->data->vector[sequenceNumber]->data[k]));
			}

			iterator++;

		}
	}

	/*  Store the newly created input structure in the master's parameter structure  */
	params->dataRho3 = tempVector;

	/*  Normal Exit  */
	DETATCHSTATUSPTR (status);
	RETURN(status);
}

